use std::{
    sync::{Arc, RwLock},
    time::Duration,
};

use iced::{
    Alignment::Center,
    Element, Task, Theme,
    widget::{button, column, row},
};
use protocol::{BUFFER_LENGTH, Error};
use tokio::{
    sync::{
        broadcast::{self},
        oneshot,
    },
    time::sleep,
};

use crate::{ClientConfig, stream::connect, voice::record};

#[derive(Debug, Default)]
struct Signal {
    microphone: Option<oneshot::Sender<bool>>,
    // speaker: Option<oneshot::Sender<bool>>,
    connection: Option<oneshot::Sender<bool>>,
}

impl Signal {
    fn reset_microphone(&mut self) -> Result<(), Error> {
        if let Some(microphone_signal) = &self.microphone {
            if !microphone_signal.is_closed() {
                self.microphone.take().expect("Never").send(true).unwrap();
                self.microphone = None;
                return Ok(());
            }
        }
        Err(Error::Signal("Reset".to_string()))
    }

    fn reset_connection(&mut self) -> Result<(), Error> {
        if let Some(connection_signal) = &self.connection {
            if !connection_signal.is_closed() {
                self.connection.take().expect("Never").send(true).unwrap();
                self.connection = None;
                return Ok(());
            }
        }
        Err(Error::Signal("Reset".to_string()))
    }
}

#[derive(Debug)]
struct Channel {
    microphone: Arc<broadcast::Sender<f32>>,
    // speaker: (broadcast::Sender<f32>, broadcast::Receiver<f32>),
}

impl Channel {
    fn new() -> Self {
        Self {
            microphone: broadcast::channel(BUFFER_LENGTH).0.into(),
            // speaker: broadcast::channel(BUFFER_LENGTH),
        }
    }
}

#[derive(Debug, Default, Clone, Copy)]
struct GUIStatus {
    room: State,
    microphone: State,
}

#[derive(Debug, Clone, Copy)]
pub enum State {
    Active,
    Passive,
    Loading,
}
impl Default for State {
    fn default() -> Self {
        State::Passive
    }
}

#[derive(Debug, Clone, Copy)]
pub enum Message {
    None,
    JoinRoom,
    LeaveRoom,
    UnmuteMicrophone,
    MuteMicrophone,
}

#[derive(Debug)]
pub struct App {
    client_config: Arc<ClientConfig>,
    gui_status: Arc<RwLock<GUIStatus>>,
    channel: Channel,
    signal: Signal,
}

impl App {
    fn reset_microphone(&mut self) -> Result<(), Error> {
        self.signal.reset_microphone()?;
        self.gui_status.write().unwrap().microphone = State::Passive;
        Ok(())
    }

    fn reset_connection(&mut self) -> Result<(), Error> {
        self.signal.reset_connection()?;
        self.gui_status.write().unwrap().room = State::Passive;
        Ok(())
    }

    pub fn theme(&self) -> Theme {
        Theme::Dark
    }

    pub fn new() -> Self {
        App {
            client_config: ClientConfig::new().into(),
            gui_status: RwLock::new(GUIStatus::default()).into(),
            channel: Channel::new(),
            signal: Signal::default(),
        }
    }
    pub fn view(&self) -> Element<'_, Message> {
        let join_room_button = match self.gui_status.read().unwrap().room {
            State::Active => button("Leave Room").on_press(Message::LeaveRoom),
            State::Passive => button("Join Room").on_press(Message::JoinRoom),
            State::Loading => button("Loading"),
        };

        let microphone_button = match self.gui_status.read().unwrap().microphone {
            State::Active => button("Mute").on_press(Message::MuteMicrophone),
            State::Passive => button("Unmute").on_press(Message::UnmuteMicrophone),
            State::Loading => button("Loading"),
        };
        column![
            row![join_room_button, microphone_button]
                .spacing(20)
                .align_y(Center)
        ]
        .align_x(Center)
        .into()
    }
    pub fn update(&mut self, message: Message) -> Task<Message> {
        match message {
            Message::None => Task::none(),
            Message::JoinRoom => {
                self.gui_status.write().unwrap().room = State::Loading;
                let client_config = self.client_config.clone();
                let gui_status = self.gui_status.clone();
                let microphone_sender_for_producing_receiver = self.channel.microphone.clone();

                let connection_signal = oneshot::channel();
                self.signal.connection = Some(connection_signal.0);
                let is_connection_started_signal = oneshot::channel();
                tokio::spawn(connect(
                    connection_signal.1,
                    is_connection_started_signal.0,
                    microphone_sender_for_producing_receiver,
                    client_config,
                ));

                let is_connection_started_task = tokio::spawn(is_connection_started_signal.1);
                Task::perform(
                    async move {
                        match is_connection_started_task.await {
                            Ok(_) => gui_status.write().unwrap().room = State::Active,
                            Err(err_val) => {
                                gui_status.write().unwrap().room = State::Passive;
                                eprintln!("Error: Connection Task | {}", err_val);
                            }
                        }
                    },
                    |_| {},
                )
                .map(|_| Message::None)
            }
            Message::LeaveRoom => {
                self.gui_status.write().unwrap().room = State::Loading;
                if let Err(err_val) = self.reset_connection() {
                    eprintln!("Error: Leave Room | {}", err_val);
                }
                Task::none()
            }
            Message::UnmuteMicrophone => {
                self.gui_status.write().unwrap().microphone = State::Loading;
                let microphone_sender = self.channel.microphone.clone();
                let microphone_stop_signal = oneshot::channel();
                self.signal.microphone = Some(microphone_stop_signal.0);
                let is_microphone_started_signal = oneshot::channel();
                tokio::spawn(record(
                    microphone_sender,
                    is_microphone_started_signal.0,
                    microphone_stop_signal.1,
                ));
                let gui_status = self.gui_status.clone();
                Task::perform(
                    async move {
                        if let Ok(_) = is_microphone_started_signal.1.await {
                            gui_status.write().unwrap().microphone = State::Active;
                        }
                    },
                    |_| Message::None,
                )
            }
            Message::MuteMicrophone => {
                self.gui_status.write().unwrap().microphone = State::Loading;
                if let Err(err_val) = self.reset_microphone() {
                    eprintln!("Error: Mute Microphone | {}", err_val);
                }
                Task::none()
            }
        }
    }
}
