use std::sync::Arc;

use cpal::traits::{DeviceTrait, HostTrait, StreamTrait};
use tokio::sync::{broadcast, oneshot};

pub async fn record(
    microphone_sender: Arc<broadcast::Sender<f32>>,
    is_microphone_started_signal: oneshot::Sender<bool>,
    microphone_stop_signal_receiver: oneshot::Receiver<bool>,
) {
    let host = cpal::default_host();
    let input_device = host.default_input_device().unwrap();
    let config = input_device.default_input_config().unwrap().into();

    let input = move |data: &[f32], _: &cpal::InputCallbackInfo| {
        for &sample in data {
            if microphone_sender.receiver_count() > 0 && sample != 0.0 {
                if let Err(err_val) = microphone_sender.send(sample) {
                    eprintln!("Error: Microphone Send | {}", err_val);
                }
            }
        }
    };

    let input_stream = input_device
        .build_input_stream(&config, input, voice_error, None)
        .unwrap();
    input_stream.play().unwrap();
    println!("Recording Started");
    if let Err(_) = is_microphone_started_signal.send(true) {
        eprintln!("Error: Is Microphone Started | Send");
    }

    tokio::task::block_in_place(|| {
        if let Err(err_val) = microphone_stop_signal_receiver.blocking_recv() {
            eprintln!(
                "Error: Microphone Stop Signal | Local Channel | {}",
                err_val
            );
        }
    });

    input_stream.pause().unwrap();
}

pub async fn play(
    mut speaker_receiver: broadcast::Receiver<f32>,
    speaker_stop_signal_receiver: oneshot::Receiver<bool>,
) {
    let host = cpal::default_host();
    let output_device = host.default_output_device().unwrap();
    let config = output_device.default_output_config().unwrap().into();

    let output = move |data: &mut [f32], _: &cpal::OutputCallbackInfo| {
        for sample in data {
            if speaker_receiver.len() > 0 {
                match speaker_receiver.blocking_recv() {
                    Ok(received_sample) => *sample = received_sample,
                    Err(err_val) => match err_val {
                        broadcast::error::RecvError::Closed => {
                            eprintln!("Error: Speaker Receive | Local Channel | Channel Closed");
                            return;
                        }
                        broadcast::error::RecvError::Lagged(lag_amount) => {
                            eprintln!(
                                "Error: Speaker Receive | Local Channel | Lagging by -> {}",
                                lag_amount
                            );
                            speaker_receiver = speaker_receiver.resubscribe();
                        }
                    },
                }
            } else {
                *sample = 0.0;
            }
        }
    };

    let output_stream = output_device
        .build_output_stream(&config, output, voice_error, None)
        .unwrap();

    output_stream.play().unwrap();
    println!("Playing Started");

    tokio::task::block_in_place(|| {
        if let Err(err_val) = speaker_stop_signal_receiver.blocking_recv() {
            eprintln!(
                "Error: Speaker Stop Signal Receive | Local Channel | {}",
                err_val
            );
        }
    });

    output_stream.pause().unwrap();
}

fn voice_error(err_val: cpal::StreamError) {
    eprintln!("Error: Voice Error | {}", err_val);
}
